"""
Class for segregated graphs (SGs).
"""
import copy
import logging

from .graph import Graph

logger = logging.getLogger(__name__)


class SG(Graph):
    def __init__(
        self,
        vertices=[],
        di_edges=set(),
        bi_edges=set(),
        ud_edges=set(),
        **kwargs,
    ):
        """
        Constructor

        :param vertices: iterable of names of vertices.
        :param di_edges: iterable of tuples of directed edges i.e. (X, Y) = X -> Y.
        :param bi_edges: iterable of tuples of bidirected edges i.e. (X, Y) = X <-> Y.
        :param ud_edges: iterable of tuples of undirected edges i.e. (X, Y) = X - Y.
        """

        # initialize vertices in SG
        super().__init__(
            vertices,
            di_edges=di_edges,
            bi_edges=bi_edges,
            ud_edges=ud_edges,
            **kwargs,
        )
        if not self._segregated():
            raise TypeError("Graph is not segregated")
        if not self._acyclic():
            raise TypeError("TypeError: Graph is not acyclic")

        # a mapping of vertices to their district ids
        # and a list of districts
        self._district_map = {}
        self._districts = []

        # a mapping of vertices to their block ids
        # and a list of blocks
        # NOTE: in an SG, only blocks of size >= 2 are considered
        self._block_map = {}
        self._blocks = []

    #### VALID GRAPH CHECKS ####
    def _acyclic(self):
        """
        Checks if the graph is directed and partially directed cycle free.

        :return: boolean indicator whether graph is acyclic.
        """

        # TODO: check correctness

        for vertex in self.vertices.values():
            if self._dfs_cycle(vertex):
                return False
        return True

    def _dfs_cycle(self, vertex):
        """
        Check if there exists a cyclic path starting at a given vertex.

        :param vertex: vertex object to start DFS at.
        :return: boolean indicating whether there is a cycle.
        """

        visit_stack = [vertex]
        prev_vertex = None
        partially_directed = False
        visited = set()

        # DFS through children and neighbors for partially directed cycle
        while visit_stack and not visited.issuperset(visit_stack):

            visited.update(visit_stack)
            v = visit_stack.pop()
            if len(v.children) != 0:
                partially_directed = True

            visit_stack.extend(
                s for s in v.children.union(v.neighbors) - {prev_vertex}
            )
            prev_vertex = v

            # if we arrive at the same vertex, then we have a cycle
            if vertex in visit_stack and partially_directed:
                return True

        return False

    def _segregated(self):
        """
        Checks if graph is segregated i.e. lacks Z <-> X - Y.

        :return: boolean indicator whether graph is segregated.
        """

        # TODO: is this enough to check segregated property?
        for vertex in self.vertices.values():
            if len(vertex.siblings) > 0 and len(vertex.neighbors) > 0:
                return False
        return True

    #### DISTRICT CODE ####
    @property
    def districts(self):
        """
        Returns list of all districts in the graph.

        :return: list of sets corresponding to districts in the graph.
        """
        return self._calculate_districts()

    def _calculate_districts(self):
        """
        Update districts in the graph.

        :return: List of districts.
        """

        self._district_map = {}
        district_counter = 0

        # Add all vertices to the district_map
        for vertex in self.vertices:
            if (
                vertex not in self._district_map
                and not self.vertices[vertex].fixed
            ):
                self._dfs_district(self.vertices[vertex], district_counter)
                district_counter += 1

        # Now process the district_map into a list of lists
        districts = [set() for _ in range(district_counter)]
        for vertex, district_id in self._district_map.items():
            districts[district_id].add(vertex)
        self._districts = districts
        return districts

    def _dfs_district(self, vertex, district_id):
        """
        DFS from vertex to discover its district.

        :param vertex: vertex object to start the DFS at.
        :param district_id: int corresponding to district ID

        :return: None
        """

        visit_stack = [vertex]
        while visit_stack:
            v = visit_stack.pop()
            self._district_map[v.name] = district_id
            visit_stack.extend(
                s for s in v.siblings if s.name not in self._district_map
            )

    def district(self, vertex):
        """
        Returns the district of a vertex.

        :param vertex: name of the vertex.
        :return: set corresponding to district.
        """

        if not self._districts:
            self.districts
        return self._districts[self._district_map[vertex]]

    #### BLOCK CODE ####
    @property
    def blocks(self):
        """
        Returns list of all blocks in the graph.

        :return: list of sets corresponding to blocks in the graph.
        """
        return self._calculate_blocks()

    def _calculate_blocks(self):
        """
        Update blocks in the graph.

        :return: None.
        """

        self._block_map = {}
        block_counter = 0

        # Add all vertices to the district_map
        for vertex in self.vertices:
            if (
                vertex not in self._block_map
                and not self.vertices[vertex].fixed
            ):
                self._dfs_block(self.vertices[vertex], block_counter)
                block_counter += 1

        # Now process the district_map into a list of lists
        self._blocks = [set() for _ in range(block_counter)]
        for vertex, block_id in self._block_map.items():
            self._blocks[block_id].add(vertex)

        return self._blocks

    def _dfs_block(self, vertex, block_id):
        """
        DFS from vertex to discover its block.

        :param vertex: vertex object to start the DFS at.
        :param block_id: int corresponding to block ID.

        :return: None
        """

        visit_stack = [vertex]
        while visit_stack:
            v = visit_stack.pop()
            self._block_map[v.name] = block_id
            visit_stack.extend(
                n for n in v.neighbors if n.name not in self._block_map
            )

    def block(self, vertex):
        """
        Returns the block of a vertex.

        :param vertex: name of the vertex.
        :return: set corresponding to block.
        """
        if not self._blocks:
            self.blocks
        block = self._blocks[self._block_map[vertex]]
        return block

    def add_biedge(self, sib1, sib2, recompute=True):
        """
        Add a bidirected edge to the graph. Overridden to recompute districts.

        :param sib1: endpoint 1 of edge.
        :param sib2: endpoint 2 of edge.
        :param recompute: boolean indicating whether districts should be recomputed.
        :return: None.
        """

        super().add_biedge(sib1, sib2)
        if recompute:
            self._districts = self._calculate_districts()

    def delete_biedge(self, sib1, sib2, recompute=True):
        """
        Delete given bidirected edge from the graph. Overridden to recompute districts.

        :param sib1: endpoint 1 of edge.
        :param sib2: endpoint 2 of edge.
        :param recompute: boolean indicating whether districts should be recomputed.
        :return: None.
        """

        super().delete_biedge(sib1, sib2)
        if recompute:
            self._districts = self._calculate_districts()

    def add_udedge(self, neb1, neb2, recompute=True):
        """
        Add an undirected edge to the graph. Overridden to recompute blocks

        :param neb1: endpoint 1 of edge.
        :param neb2: endpoint 2 of edge.
        :param recompute: boolean indicating whether blocks should be recomputed.
        :return: None.
        """

        super().add_udedge(neb1, neb2)
        if recompute:
            self._blocks = self._calculate_blocks()

    def delete_udedge(self, neb1, neb2, recompute=True):
        """
        Delete given undirected edge from the graph. Overridden to recompute blocks.

        :param neb1: endpoint 1 of edge.
        :param neb2: endpoint 2 of edge.
        :param recompute: boolean indicating whether blocks should be recomputed.
        :return: None.
        """

        super().delete_udedge(neb1, neb2)
        if recompute:
            self._blocks = self._calculate_blocks()

    def fix(self, vertices):
        """
        Perform the graphical operation of fixing on a set of vertices.

        Does not check that vertices are fixable.

        :param vertices: iterable of vertices to be fixed.
        :return: None.
        """
        if type(vertices) == str:
            vertices = [vertices]

        for v in vertices:

            self.vertices[v].fixed = True

            # delete incoming directed edges
            parents = [p.name for p in self.vertices[v].parents]
            for p in parents:
                self.delete_diedge(p, v)

            # delete bidirected edges
            siblings = [s.name for s in self.vertices[v].siblings]
            for s in siblings:
                self.delete_biedge(s, v, recompute=False)

            # delete undirected edges
            neighbors = [n.name for n in self.vertices[v].neighbors]
            for n in neighbors:
                if self.vertices[n].fixed:
                    self.delete_udedge(n, v, recompute=False)

        # recompute the districts and blocks as they may have changed
        self._districts = self._calculate_districts()
        self._blocks = self._calculate_blocks()

        return self

    def fixable(self, vertices):
        """
        Check if there exists a valid fixing order and return such
        an order in the form of a list, else returns an empty list.

        :param vertices: set of vertices to check fixability for.
        :return: a boolean indicating whether the set was fixable and a valid fixing order as a stack.
        """

        # keep track of vertices still left to fix
        # and initialize a fixing order
        G = copy.deepcopy(self)
        remaining_vertices = list(sorted(vertices))
        fixing_order = []
        fixed = True  # flag to check if we fixed a variable on each pass

        # while we have more vertices to fix, and were able to perform a fix
        while remaining_vertices and fixed:

            fixed = False

            for v in remaining_vertices:

                # check if any nodes are reachable via -> AND <->
                # by looking at intersection of district and descendants
                if len(G.descendants([v]).intersection(G.district(v))) == 1:
                    G.fix([v])
                    remaining_vertices.remove(v)
                    fixing_order.append(v)
                    fixed = True
                    break

            # if unsuccessful, return failure and
            # fixing order up until point of failure
            if not fixed:
                return False, fixing_order

        # if fixing vertices was successful, return success
        # and the fixing order
        return True, fixing_order



"""
Class that provides an interface to estimation strategies for the counterfactual mean E[Y(t)]
"""

import copy

import numpy as np
import statsmodels.api as sm
from scipy.stats import norm

from ananke.identification import OneLineID


class CausalEffect:
    """
    Provides an interface to various estimation strategies for the ACE: E[Y(1) - Y(0)].
    """

    def __init__(self, graph, treatment, outcome):
        """
        Constructor.

        :param graph: ADMG corresponding to substantive knowledge/model.
        :param treatment: name of vertex corresponding to the treatment.
        :param outcome: name of vertex corresponding to the outcome.
        """

        self.graph = copy.deepcopy(graph)
        self.treatment = treatment
        self.outcome = outcome
        self.strategy = None
        self.is_mb_shielded = self.graph.mb_shielded()

        # a dictionary of names for available estimators
        self.estimators = {
            "ipw": self._ipw,
            "gformula": self._gformula,
            "aipw": self._aipw,
            "eff-aipw": self._eff_augmented_ipw,
            "p-ipw": self._primal_ipw,
            "d-ipw": self._dual_ipw,
            "apipw": self._augmented_primal_ipw,
            "eff-apipw": self._eff_augmented_primal_ipw,
            "n-ipw": self._nested_ipw,
            "anipw": self._augmented_nested_ipw,
        }

        # a dictionary of names for available modeling strategies
        self.models = {
            "glm-binary": self._fit_binary_glm,
            "glm-continuous": self._fit_continuous_glm,
        }

        # check if the query is ID
        self.one_id = OneLineID(graph, [treatment], [outcome])

        # check the fixability criteria for the treatment
        if (
            len(
                self.graph.district(treatment).intersection(
                    self.graph.descendants([treatment])
                )
            )
            == 1
        ):
            self.strategy = "a-fixable"
            if self.is_mb_shielded:
                print(
                    "\n Treatment is a-fixable and graph is mb-shielded. \n\n Available estimators are:\n \n"
                    + "1. IPW (ipw)\n"
                    + "2. Outcome regression (gformula)\n"
                    + "3. Generalized AIPW (aipw)\n"
                    + "4. Efficient Generalized AIPW (eff-aipw) \n \n"
                    + "Suggested estimator is Efficient Generalized AIPW \n"
                )
            else:
                print(
                    "\n Treatment is a-fixable.\n\n Available estimators are :\n \n"
                    + "1. IPW (ipw)\n"
                    + "2. Outcome regression (gformula)\n"
                    + "3. Generalized AIPW (aipw)\n \n"
                    + "Suggested estimator is Generalized AIPW \n"
                )

        elif (
            len(
                self.graph.district(treatment).intersection(
                    self.graph.children([treatment])
                )
            )
            == 0
        ):
            self.strategy = "p-fixable"
            if self.is_mb_shielded:
                print(
                    "\n Treatment is p-fixable and graph is mb-shielded. \n\n Available estimators are:\n\n"
                    + "1. Primal IPW (p-ipw)\n"
                    + "2. Dual IPW (d-ipw)\n"
                    + "3. APIPW (apipw)\n"
                    + "4. Efficient APIPW (eff-apipw) \n \n"
                    + "Suggested estimator is Efficient APIPW \n"
                )
            else:
                print(
                    "\n Treatment is p-fixable. \n\n Available estimators are:\n \n"
                    + "1. Primal IPW (p-ipw)\n"
                    + "2. Dual IPW (d-ipw)\n"
                    + "3. APIPW (apipw) \n\n"
                    + "Suggested estimator is APIPW \n"
                )

        elif self.one_id.id():
            self.strategy = "nested-fixable"
            print(
                "\n Effect is identified. \n \n Available estimators:\n \n"
                + "1. Nested IPW (n-ipw)\n"
                + "2. Augmented NIPW (anipw) \n\n"
                + "Suggested estimator is Augmented NIPW \n"
            )

        else:
            self.strategy = "Not ID"
            print("Effect is not identified! \n")

        # finally, fix a valid topological order for estimation queries
        self.p_order = self._find_valid_order(
            "p-fixable"
        )  # used for a-fixable/p-fixable strategies
        self.n_order = self._find_valid_order(
            "nested-fixable"
        )  # used for nested-fixable strategies
        self.state_space_map_ = (
            {}
        )  # maps from variable names to state space of the variable (binary/continuous)

    def _find_valid_order(self, order_type):
        """
        Utility function to find a valid order that satisfies the requirements in
        Semiparametric Inference (Bhattacharya, Nabi, & Shpitser 2020) paper.

        :param order_type: string specifying if the topological order is required for p-fixable or n-fixable problems.
        :return: list corresponding to a valid topological order.
        """

        # if we are a-fixing or p-fixing, just ensure that the
        # treatment appears after all its non-descendants
        if order_type != "nested-fixable":
            focal_vertices = {self.treatment}

        # otherwise we ensure the vertices in Y* that intersect with district
        # of the treatment appear after all their non-descendants
        else:
            focal_vertices = self.one_id.ystar.intersection(
                self.graph.district(self.treatment)
            )

        # perform a sort on the nondescendants first
        nondescendants = set(self.graph.vertices) - self.graph.descendants(
            focal_vertices
        )
        order = self.graph.subgraph(nondescendants).topological_sort()

        # then sort the focal vertices and its descendants
        order += self.graph.subgraph(
            self.graph.descendants(focal_vertices)
        ).topological_sort()

        return order

    def _fit_binary_glm(self, data, formula, weights=None):
        """
        Fit a binary GLM to data given a formula.

        :param data: pandas data frame containing the data.
        :param formula: string encoding an R-style formula e.g: Y ~ X1 + X2.
        :return: the fitted model.
        """

        return sm.GLM.from_formula(
            formula,
            data=data,
            family=sm.families.Binomial(),
            freq_weights=weights,
        ).fit()

    def _fit_continuous_glm(self, data, formula, weights=None):
        """
        Fit a linear GLM to data given a formula.

        :param data: pandas data frame containing the data.
        :param formula: string encoding an R-style formula: e.g. Y ~ X1 + X2.
        :return: the fitted model.
        """

        return sm.GLM.from_formula(
            formula,
            data=data,
            family=sm.families.Gaussian(),
            freq_weights=weights,
        ).fit()

    def _ipw(self, data, assignment, model_binary=None, model_continuous=None):
        """
        IPW estimator for the counterfactual mean E[Y(t)].

        :param data: pandas data frame containing the data.
        :param assignment: assignment value for treatment.
        :param model_binary: string specifying modeling strategy to use for propensity score: e.g. glm-binary.
        :param model_continuous: this argument is ignored for IPW.
        :return: float corresponding to computed E[Y(t)].
        """

        # pedantic checks to make sure the method returns valid estimates
        if self.strategy != "a-fixable":
            raise RuntimeError(
                "IPW will not return valid estimates as treatment is not a-fixable"
            )

        # extract outcome from data frame and compute Markov pillow of treatment
        Y = data[self.outcome]
        mp_T = self.graph.markov_pillow([self.treatment], self.p_order)

        if len(mp_T) != 0:
            # fit T | mp(T) and compute probability of treatment for each sample
            formula = self.treatment + " ~ " + "+".join(mp_T)  # + "+ ones"
            model = model_binary(data, formula)
            prob_T = model.predict(data)
        else:
            prob_T = np.ones(len(data)) * np.mean(data[self.treatment])

        indices_T0 = data.index[data[self.treatment] == 0]
        prob_T[indices_T0] = 1 - prob_T[indices_T0]

        # compute IPW estimate
        indices = data[self.treatment] == assignment
        return np.mean((indices / prob_T) * Y)

    def _gformula(
        self, data, assignment, model_binary=None, model_continuous=None
    ):
        """
        Outcome regression estimator for the counterfactual mean E[Y(t)].

        :param data: pandas data frame containing the data.
        :param assignment: assignment value for treatment.
        :param model_binary: string specifying modeling strategy to use for binary variables: e.g. glm-binary.
        :param model_continuous: string specifying modeling strategy to use for outcome regression: e.g. glm-continuous.
        :return: float corresponding to computed E[Y(t)].
        """

        # pedantic checks to make sure the method returns valid estimates
        if self.strategy != "a-fixable":
            raise RuntimeError(
                "g-formula will not return valid estimates as treatment is not a-fixable"
            )

        # create a dataset where T=t
        data_assign = data.copy()
        data_assign[self.treatment] = assignment

        # fit Y | T=t, mp(T)
        mp_T = self.graph.markov_pillow([self.treatment], self.p_order)
        if len(mp_T) != 0:
            formula = (
                self.outcome + " ~ " + self.treatment + "+" + "+".join(mp_T)
            )
            # predict outcome appropriately depending on binary vs continuous
        else:
            formula = self.outcome + " ~ " + self.treatment

        if self.state_space_map_[self.outcome] == "binary":
            model = model_binary(data, formula)
        else:
            model = model_continuous(data, formula)

        return np.mean(model.predict(data_assign))

    def _aipw(self, data, assignment, model_binary=None, model_continuous=None):
        """
        Augmented IPW estimator for the counterfactual mean E[Y(t)].

        :param data: pandas data frame containing the data.
        :param assignment: assignment value for treatment.
        :param model_binary: string specifying modeling strategy to use for binary variables: e.g. glm-binary.
        :param model_continuous: string specifying modeling strategy to use for continuous variables: e.g. glm-continuous.
        :return: float corresponding to computed E[Y(t)].
        """

        # pedantic checks to make sure the method returns valid estimates
        if self.strategy != "a-fixable":
            raise RuntimeError(
                "Augmented IPW will not return valid estimates as treatment is not a-fixable"
            )

        # extract the outcome and get Markov pillow of the treatment
        Y = data[self.outcome]
        mp_T = self.graph.markov_pillow([self.treatment], self.p_order)

        if len(mp_T) != 0:
            # fit T | mp(T) and predict treatment probabilities
            formula_T = self.treatment + " ~ " + "+".join(mp_T)  # + "+ ones"
            model = model_binary(data, formula_T)
            prob_T = model.predict(data)
            formula_Y = (
                self.outcome + " ~ " + self.treatment + "+" + "+".join(mp_T)
            )
        else:
            prob_T = np.ones(len(data)) * np.mean(data[self.treatment])
            formula_Y = self.outcome + " ~ " + self.treatment

        indices_T0 = data.index[data[self.treatment] == 0]
        prob_T[indices_T0] = 1 - prob_T[indices_T0]
        indices = data[self.treatment] == assignment

        # fit Y | T=t, mp(T) and predict outcomes under assignment T=t
        data_assign = data.copy()
        data_assign[self.treatment] = assignment
        if self.state_space_map_[self.outcome] == "binary":
            model = model_binary(data, formula_Y)
        else:
            model = model_continuous(data, formula_Y)
        Yhat_vec = model.predict(data_assign)

        # return AIPW estimate
        return np.mean((indices / prob_T) * (Y - Yhat_vec) + Yhat_vec)

    def _eff_augmented_ipw(
        self, data, assignment, model_binary=None, model_continuous=None
    ):
        """
        Efficient augmented IPW estimator for the counterfactual mean E[Y(t)].

        :param data: pandas data frame containing the data.
        :param assignment: assignment value for treatment.
        :param model_binary: string specifying modeling strategy to use for binary variables: e.g. glm-binary.
        :param model_continuous: string specifying modeling strategy to use for continuous variables: e.g. glm-continuous.
        :return: float corresponding to computed E[Y(t)].
        """

        # pedantic checks to make sure the method returns valid estimates
        if self.strategy != "a-fixable":
            raise RuntimeError(
                "Augmented IPW will not return valid estimates as treatment is not a-fixable"
            )
        if not self.is_mb_shielded:
            raise RuntimeError(
                "EIF will not return valid estimates as graph is not mb-shielded"
            )

        # extract the outcome and get Markov pillow of treatment
        Y = data[self.outcome]
        mp_T = self.graph.markov_pillow([self.treatment], self.p_order)

        # fit T | mp(T) and compute treatment probabilities
        if len(mp_T) != 0:
            formula = self.treatment + " ~ " + "+".join(mp_T)  # + "+ ones"
            model = model_binary(data, formula)
            prob_T = model.predict(data)
        else:
            prob_T = np.ones(len(data)) * np.mean(data[self.treatment])

        indices_T0 = data.index[data[self.treatment] == 0]
        prob_T[indices_T0] = 1 - prob_T[indices_T0]
        indices = data[self.treatment] == assignment

        # compute the primal estimates that we use to compute projections
        primal = (indices / prob_T) * Y
        data["primal"] = primal
        eif_vec = 0

        # get the variables that are actually involved in the efficient influence function
        # TODO: prune extra variables
        eif_vars = [V for V in self.graph.vertices]
        eif_vars.remove(self.treatment)

        # iterate over variables
        for V in eif_vars:

            # get the Markov pillow of the variable
            mpV = self.graph.markov_pillow([V], self.p_order)

            # compute E[primal | mp(V)]
            formula = "primal ~ " + "+".join(mpV)

            # special case if mp(V) is empty, then E[primal | mp(V)] = E[primal]
            if len(mpV) != 0:
                model_mpV = model_continuous(
                    data, formula
                )  # primal is a continuous r.v.
                primal_mpV = model_mpV.predict(data)
            else:
                primal_mpV = np.mean(primal)

            # compute E[primal | V, mp(V)]
            formula = formula + "+" + V
            model_VmpV = model_continuous(data, formula)
            primal_VmpV = model_VmpV.predict(data)

            # add contribution of current variable: E[primal | V, mp(V)] - E[primal | mp(V)]
            eif_vec += primal_VmpV - primal_mpV

        # re-add the primal so final result is not mean zero
        eif_vec = eif_vec + np.mean(primal)

        # return efficient AIPW estimate
        return np.mean(eif_vec)

    def _beta_primal(
        self, data, assignment, model_binary=None, model_continuous=None
    ):
        """
        Utility function to compute primal estimates for a dataset.

        :param data: pandas data frame containing the data.
        :param assignment: assignment value for treatment.
        :param model_binary: string specifying modeling strategy to use for binary variables: e.g. glm-binary.
        :param model_continuous: string specifying modeling strategy to use for continuous variables: e.g. glm-continuous.
        :return: numpy array of floats corresponding to primal estimates for each sample.
        """

        # extract the outcome
        Y = data[self.outcome]

        # C := pre-treatment vars and L := post-treatment vars in district of treatment
        C = self.graph.pre([self.treatment], self.p_order)
        post = set(self.graph.vertices).difference(C)
        L = post.intersection(self.graph.district(self.treatment))

        # create copies of the data with treatment assignments T=0 and T=1
        data_T1 = data.copy()
        data_T1[self.treatment] = 1
        data_T0 = data.copy()
        data_T0[self.treatment] = 0

        indices = data[self.treatment] == assignment

        # prob: stores \prod_{Li in L} p(Li | mp(Li))
        # prob_T1: stores \prod_{Li in L} p(Li | mp(Li)) at T=1
        # prob_T0: stores \prod_{Li in L} p(Li | mp(Li)) at T=0

        mp_T = self.graph.markov_pillow([self.treatment], self.p_order)
        indices_T0 = data.index[data[self.treatment] == 0]

        if len(mp_T) != 0:
            formula = self.treatment + " ~ " + "+".join(mp_T)
            model = model_binary(data, formula)
            prob = model.predict(data)
            prob[indices_T0] = 1 - prob[indices_T0]
            prob_T1 = model.predict(data)
            prob_T0 = 1 - prob_T1
        else:
            prob = np.ones(len(data)) * np.mean(data[self.treatment])
            prob[indices_T0] = 1 - prob[indices_T0]
            prob_T1 = np.ones(len(data)) * np.mean(data[self.treatment])
            prob_T0 = 1 - prob_T1

        # iterate over vertices in L (except the outcome)
        for V in L.difference([self.treatment, self.outcome]):

            # fit V | mp(V)
            mp_V = self.graph.markov_pillow([V], self.p_order)
            formula = V + " ~ " + "+".join(mp_V)

            # p(V =v | .), p(V = v | . , T=1), p(V = v | ., T=0)
            if self.state_space_map_[V] == "binary":
                model = model_binary(data, formula)
                prob_V = model.predict(data)
                prob_V_T1 = model.predict(data_T1)
                prob_V_T0 = model.predict(data_T0)

                indices_V0 = data.index[data[V] == 0]

                # p(V | .), p(V | ., T=t)
                prob_V[indices_V0] = 1 - prob_V[indices_V0]
                prob_V_T1[indices_V0] = 1 - prob_V_T1[indices_V0]
                prob_V_T0[indices_V0] = 1 - prob_V_T0[indices_V0]

            else:
                model = model_continuous(data, formula)
                E_V = model.predict(data)
                E_V_T1 = model.predict(data_T1)
                E_V_T0 = model.predict(data_T0)

                std = np.std(data[V] - E_V)
                prob_V = norm.pdf(data[V], loc=E_V, scale=std)
                prob_V_T1 = norm.pdf(data[V], loc=E_V_T1, scale=std)
                prob_V_T0 = norm.pdf(data[V], loc=E_V_T0, scale=std)

            prob *= prob_V
            prob_T1 *= prob_V_T1
            prob_T0 *= prob_V_T0

        # special case when the outcome is in L
        if self.outcome in L:

            # fit a binary/continuous model for Y | mp(Y)
            mp_Y = self.graph.markov_pillow([self.outcome], self.p_order)
            formula = self.outcome + " ~ " + "+".join(mp_Y)
            if self.state_space_map_[self.outcome] == "binary":
                model = model_binary(data, formula)
            else:
                model = model_continuous(data, formula)

            # predict the outcome and adjust numerator of primal accordingly
            Yhat_T1 = model.predict(data_T1)
            Yhat_T0 = model.predict(data_T0)
            prob_sumT = prob_T1 * Yhat_T1 + prob_T0 * Yhat_T0
            beta_primal = indices * (prob_sumT / prob)
        else:
            prob_sumT = prob_T1 + prob_T0
            beta_primal = indices * (prob_sumT / prob) * Y

        return beta_primal

    def _primal_ipw(
        self, data, assignment, model_binary=None, model_continuous=None
    ):
        """
        Primal IPW estimator for the counterfactual mean E[Y(t)].

        :param data: pandas data frame containing the data.
        :param assignment: assignment value for treatment.
        :param model_binary: string specifying modeling strategy to use for binary variables: e.g. glm-binary.
        :param model_continuous: string specifying modeling strategy to use for continuous variables: e.g. glm-continuous.
        :return: float corresponding to computed E[Y(t)].
        """

        # pedantic checks to make sure the method returns valid estimates
        if self.strategy != "p-fixable" and self.strategy != "a-fixable":
            raise RuntimeError(
                "Primal IPW will not return valid estimates as treatment is not p-fixable"
            )

        # return primal IPW estimate
        return np.mean(
            self._beta_primal(data, assignment, model_binary, model_continuous)
        )

    def _beta_dual(
        self, data, assignment, model_binary=None, model_continuous=None
    ):
        """
        Utility function to compute dual estimates for a dataset.

        :param data: pandas data frame containing the data.
        :param assignment: assignment value for treatment.
        :param model_binary: string specifying modeling strategy to use for binary variables: e.g. glm-binary.
        :param model_continuous: string specifying modeling strategy to use for continuous variables: e.g. glm-continuous.
        :return: numpy array of floats corresponding to dual estimates for each sample.
        """

        # extract the outcome
        Y = data[self.outcome]

        # M := inverse Markov pillow of the treatment
        M = set(
            [
                m
                for m in self.graph.vertices
                if self.treatment in self.graph.markov_pillow([m], self.p_order)
            ]
        )
        M = M.difference(self.graph.district(self.treatment))

        # create a dataset where T=t
        data_assigned = data.copy()
        data_assigned[self.treatment] = assignment

        prob = 1  # stores \prod_{Mi in M} p(Mi | mp(Mi))|T=t / p(Mi | mp(Mi))
        for V in M.difference([self.outcome]):

            # Fit V | mp(V)
            mp_V = self.graph.markov_pillow([V], self.p_order)
            formula = V + " ~ " + "+".join(mp_V)

            # p(V = 1 | .), p(V = 1 | . , T=assigned)
            if self.state_space_map_[V] == "binary":
                model = model_binary(data, formula)
                prob_V = model.predict(data)
                prob_V_assigned = model.predict(data_assigned)

                indices_V0 = data.index[data[V] == 0]

                # p(V | .) and p(V | ., T=assignment)
                prob_V[indices_V0] = 1 - prob_V[indices_V0]
                prob_V_assigned[indices_V0] = 1 - prob_V_assigned[indices_V0]

            else:
                model = model_continuous(data, formula)
                E_V = model.predict(data)
                E_V_assigned = model.predict(data_assigned)

                std = np.std(data[V] - E_V)
                prob_V = norm.pdf(data[V], loc=E_V, scale=std)
                prob_V_assigned = norm.pdf(data[V], loc=E_V_assigned, scale=std)

            prob *= prob_V_assigned / prob_V

        # special case for if the outcome is in M
        if self.outcome in M:
            mp_Y = self.graph.markov_pillow([self.outcome], self.p_order)
            formula = self.outcome + " ~ " + "+".join(mp_Y)
            if self.state_space_map_[self.outcome] == "binary":
                model = model_binary(data, formula)
            else:
                model = model_continuous(data, formula)
            Yhat_assigned = model.predict(data_assigned)
        else:
            Yhat_assigned = Y

        return prob * Yhat_assigned

    def _dual_ipw(
        self, data, assignment, model_binary=None, model_continuous=None
    ):
        """
        Dual IPW estimator for the counterfactual mean E[Y(t)].

        :param data: pandas data frame containing the data.
        :param assignment: assignment value for treatment.
        :param model_binary: string specifying modeling strategy to use for binary variables: e.g. glm-binary.
        :param model_continuous: string specifying modeling strategy to use for continuous variables: e.g. glm-continuous.
        :return: float corresponding to computed E[Y(t)].
        """

        # pedantic checks to make sure the method returns valid estimates
        if self.strategy != "p-fixable" and self.strategy != "a-fixable":
            raise RuntimeError(
                "Dual IPW will not return valid estimates as treatment is not p-fixable"
            )

        # return primal dual IPW estimate
        return np.mean(
            self._beta_dual(data, assignment, model_binary, model_continuous)
        )

    def _augmented_primal_ipw(
        self, data, assignment, model_binary=None, model_continuous=None
    ):
        """
        Augmented primal IPW estimator for the counterfactual mean E[Y(t)].

        :param data: pandas data frame containing the data.
        :param assignment: assignment value for treatment.
        :param model_binary: string specifying modeling strategy to use for binary variables: e.g. glm-binary.
        :param model_continuous: string specifying modeling strategy to use for continuous variables: e.g. glm-continuous.
        :return: float corresponding to computed E[Y(t)].
        """

        # pedantic checks to make sure the method returns valid estimates
        if self.strategy != "p-fixable" and self.strategy != "a-fixable":
            raise RuntimeError(
                "Augmented primal IPW will not return valid estimates as treatment is not p-fixable"
            )

        # compute the primal and dual estimates and them to the data frame
        beta_primal = self._beta_primal(
            data, assignment, model_binary, model_continuous
        )
        beta_dual = self._beta_dual(
            data, assignment, model_binary, model_continuous
        )
        data.loc[:, "beta_primal"] = beta_primal
        data.loc[:, "beta_dual"] = beta_dual

        # C := pre-treatment vars
        # L := post-treatment vars in the district of T
        # M := post treatment vars not in L (a.k.a. the rest)
        C = self.graph.pre([self.treatment], self.p_order)
        post = set(self.graph.vertices).difference(C)
        L = post.intersection(self.graph.district(self.treatment))
        M = post - L

        IF = 0

        # iterate over all post-treatment variables
        for V in post:

            # compute all predecessors according to the topological order
            pre_V = self.graph.pre([V], self.p_order)

            # if the variables is in M, project using the primal otherwise use the dual
            # to fit E[beta | pre(V)]
            if V in M:
                formula = "beta_primal" + " ~ " + "+".join(pre_V)
            elif V in L:
                formula = "beta_dual" + " ~ " + "+".join(pre_V)

            # special logic for if there are no predecessors for the variable (which could only happen to T)
            if len(pre_V) != 0:
                model_preV = model_continuous(
                    data, formula
                )  # primal/dual is a continuous r.v.
                pred_preV = model_preV.predict(data)
            else:
                pred_preV = 0

            # fit E[beta | V, pre(V)]
            formula = formula + " + " + V
            model_VpreV = model_continuous(data, formula)
            pred_VpreV = model_VpreV.predict(data)

            # add contribution of current variable as E[beta | V, pre(V)] - E[beta | pre(V)]
            IF += pred_VpreV - pred_preV

        # final contribution from E[beta | C] (if C is not empty)
        if len(C) != 0:
            formula = "beta_dual" + " ~ " + "+".join(C)
            model_C = model_continuous(
                data, formula
            )  # dual is a continuous r.v.
            IF += model_C.predict(data)

        # return APIPW estimate
        return np.mean(IF)

    def _eff_augmented_primal_ipw(
        self, data, assignment, model_binary=None, model_continuous=None
    ):
        """
        Efficient augmented primal IPW estimator for the counterfactual mean E[Y(t)].

        :param data: pandas data frame containing the data.
        :param assignment: assignment value for treatment.
        :param model_binary: string specifying modeling strategy to use for binary variables: e.g. glm-binary.
        :param model_continuous: string specifying modeling strategy to use for continuous variables: e.g. glm-continuous.
        :return: float corresponding to computed E[Y(t)].
        """

        # pedantic checks to make sure the method returns valid estimates
        if self.strategy != "p-fixable" and self.strategy != "a-fixable":
            raise RuntimeError(
                "Augmented primal IPW will not return valid estimates as treatment is not p-fixable"
            )
        if not self.is_mb_shielded:
            raise RuntimeError(
                "EIF will not return valid estimates as graph is not mb-shielded"
            )

        # compute primal and dual estimates and add them to the data frame
        beta_primal = self._beta_primal(
            data, assignment, model_binary, model_continuous
        )
        beta_dual = self._beta_dual(
            data, assignment, model_binary, model_continuous
        )
        data.loc[:, "beta_primal"] = beta_primal
        data.loc[:, "beta_dual"] = beta_dual

        # C := pre-treatment vars
        # L := post-treatment vars in the district of T
        # M := post treatment vars not in L (a.k.a. the rest)
        C = self.graph.pre([self.treatment], self.p_order)
        post = set(self.graph.vertices).difference(C)
        L = post.intersection(self.graph.district(self.treatment))
        M = post - L

        IF = 0

        # iterate over all variables
        for V in self.graph.vertices:

            # get the Markov pillow
            mp_V = self.graph.markov_pillow([V], self.p_order)

            # if the variables is in M, project using the primal otherwise use the dual
            # to fit E[beta | mp(V)]
            if V in M:
                formula = "beta_primal" + " ~ " + "+".join(mp_V)
            else:
                formula = "beta_dual" + " ~ " + "+".join(mp_V)

            # special logic for if there the Markov pillow is empty
            if len(mp_V) != 0:
                model_mpV = model_continuous(data, formula)
                pred_mpV = model_mpV.predict(data)
            else:
                pred_mpV = np.mean(beta_dual)

            # fit E[beta | V, mp(V)]
            formula = formula + " + " + V
            model_VmpV = model_continuous(data, formula)
            pred_VmpV = model_VmpV.predict(data)

            # add contribution of current variable as E[beta | V, mp(V)] - E[beta | mp(V)]
            IF += pred_VmpV - pred_mpV

        # add final contribution so that estimator is not mean-zero
        IF += np.mean(beta_dual)

        # return efficient APIPW estimate
        return np.mean(IF)

    def _fit_intrinsic_kernel(
        self, data, district, model_binary=None, model_continuous=None
    ):
        """
        Get estimates of an intrinsic kernel q_D(D|pa(D)) for each sample.

        :param data: pandas data frame containing the data.
        :param model_binary: string specifying modeling strategy to use for binary variables: e.g. glm-binary.
        :param model_continuous: string specifying modeling strategy to use for continuous variables: e.g. glm-continuous.
        :return: numpy array of estimated probabilities of the kernel for each sample.
        """

        fixing_prob = np.ones(len(data))

        # get parents of the district
        parents = self.graph.parents(district) - district
        # initialize with ancestral margin as fixing descendants are just sums
        G = self.graph.subgraph(self.graph.ancestors(district))

        # we know D is intrinsic because of prior checks
        remaining_vertices = (
            set(G.vertices) - district - parents
        )  # is subtracting parents here ok? think so..

        while not G.fixable(parents)[0]:

            fixed = False

            # at every step try to simplify as much as possible by fixing
            # childless vertices as these correspond to just sums
            # also find a backup fixable vertex in case there are no childless ones
            iter_gen = iter(remaining_vertices)
            i = 0

            while not fixed and i < len(remaining_vertices):
                i += 1
                V = next(iter_gen)
                if len(G.children([V])) == 0:
                    G.fix([V])
                    fixed = True
                    remaining_vertices.remove(V)
                elif G.fixable([V])[0]:
                    fixable_V = V

            # if we fixed something go back and see if we can fix another childless vertex
            # or if the requirement that parents are fixable is satisfied
            if fixed:
                continue

            # otherwise do a true fix corresponding to reweighting
            # using the fixable V that we found
            V = fixable_V
            mb_V = G.markov_blanket([V])
            formula = V + " ~ " + "+".join(mb_V)

            # p(V = 1 | .)
            if self.state_space_map_[V] == "binary":
                model = model_binary(data, formula, 1 / fixing_prob)
                prob_V = model.predict(data)
                indices_V0 = data.index[data[V] == 0]

                # p(V | .)
                prob_V[indices_V0] = 1 - prob_V[indices_V0]

            # handling continuous data
            else:
                model = model_continuous(data, formula, 1 / fixing_prob)
                E_V = model.predict(data)
                std = np.std(data[V] - E_V)
                prob_V = norm.pdf(data[V], loc=E_V, scale=std)

            fixing_prob *= prob_V
            G.fix([V])
            remaining_vertices.remove(V)

        # get the Markov blanket of the parents in this final graph
        mb_parents = G.markov_blanket(parents)

        # iterate over each parent and get weights required to fix them
        for V in parents:

            # p(V = 1 | .)
            mp_V = mb_parents.intersection(self.graph.pre([V], self.n_order))
            if len(mp_V) != 0:
                formula = V + " ~ " + "+".join(mp_V)
            else:
                formula = V + " ~ -1 + 1"

            if self.state_space_map_[V] == "binary":
                model = model_binary(data, formula, 1 / fixing_prob)
                prob_V = model.predict(data)
                indices_V0 = data.index[data[V] == 0]

                # p(V | .)
                prob_V[indices_V0] = 1 - prob_V[indices_V0]

            # handling continuous data
            else:
                model = model_continuous(data, formula, 1 / fixing_prob)
                E_V = model.predict(data)
                std = np.std(data[V] - E_V)
                prob_V = norm.pdf(data[V], loc=E_V, scale=std)

            fixing_prob *= prob_V

        kernel_prob = np.ones(len(data))

        # iterate over member of the intrinsic set and get the final kernel weights
        for V in district:

            # p(V = 1 | .)
            mp_V = parents.union(
                district.intersection(self.graph.pre(V, self.n_order))
            )
            if len(mp_V) != 0:
                formula = V + " ~ " + "+".join(mp_V)
            else:
                formula = V + " ~ -1 + 1"

            if self.state_space_map_[V] == "binary":
                model = model_binary(data, formula, 1 / fixing_prob)
                prob_V = model.predict(data)
                indices_V0 = data.index[data[V] == 0]

                # p(V | .)
                prob_V[indices_V0] = 1 - prob_V[indices_V0]

            # handling continuous data
            else:
                model = model_continuous(data, formula, 1 / fixing_prob)
                E_V = model.predict(data)
                std = np.std(data[V] - E_V)
                prob_V = norm.pdf(data[V], loc=E_V, scale=std)

            kernel_prob *= prob_V

        return kernel_prob

    def _get_nested_rebalanced_weights(
        self, data, model_binary=None, model_continuous=None
    ):
        """
        Get the rebalancing weights required for nested IPW and augmented nested IPW

        :param data: pandas data frame containing the data.
        :param model_binary: string specifying modeling strategy to use for binary variables: e.g. glm-binary.
        :param model_continuous: string specifying modeling strategy to use for continuous variables: e.g. glm-continuous.
        :return: numpy array corresponding to rebalancing weights.
        """

        # first get all districts of GY* that intersect with district of T
        district_T = self.graph.district(self.treatment)
        modified_districts = [
            district
            for district in self.one_id.Gystar.districts
            if len(district.intersection(district_T)) != 0
        ]

        # placeholder for the rebalancing probability
        rebalance_prob = np.ones(len(data))

        # iterate over each modified district
        for district in modified_districts:

            # first compute weights in the denominator of \prod_{Vi in district} p(Vi | mp(Vi))
            for V in district:

                # Fit V | mp(V)
                mp_V = self.graph.markov_pillow([V], self.n_order)
                formula = V + " ~ " + "+".join(mp_V)

                # p(V = 1 | .)
                if self.state_space_map_[V] == "binary":
                    model = model_binary(data, formula)
                    prob_V = model.predict(data)
                    indices_V0 = data.index[data[V] == 0]

                    # p(V | .)
                    prob_V[indices_V0] = 1 - prob_V[indices_V0]

                # handling continuous data
                else:
                    model = model_continuous(data, formula)
                    E_V = model.predict(data)
                    std = np.std(data[V] - E_V)
                    prob_V = norm.pdf(data[V], loc=E_V, scale=std)

                rebalance_prob *= 1 / prob_V

            # now compute the q_D(D | pa(D))
            rebalance_prob *= self._fit_intrinsic_kernel(
                data, district, model_binary, model_continuous
            )

        return 1 / rebalance_prob

    def _nested_ipw(
        self, data, assignment, model_binary=None, model_continuous=None
    ):
        """
        Nested IPW estimator for the counterfactual mean E[Y(t)].

        :param data: pandas data frame containing the data.
        :param assignment: assignment value for treatment.
        :param model_binary: string specifying modeling strategy to use for binary variables: e.g. glm-binary.
        :param model_continuous: string specifying modeling strategy to use for continuous variables: e.g. glm-continuous.
        :return: float corresponding to computed E[Y(t)].
        """

        # pedantic checks to make sure the method returns valid estimates
        if self.strategy == "Not ID":
            raise RuntimeError(
                "Nested IPW will not return valid estimates as causal effect is not identified"
            )

        # fit T | mp(T) with the rebalanced weights and compute the nested IPW
        rebalance_weights = self._get_nested_rebalanced_weights(
            data, model_binary, model_continuous
        )
        # extract outcome from data frame and compute Markov pillow of treatment
        Y = data[self.outcome]
        mp_T = self.graph.markov_pillow([self.treatment], self.n_order)

        if len(mp_T) != 0:
            # fit T | mp(T) and compute probability of treatment for each sample
            formula = self.treatment + " ~ " + "+".join(mp_T)
            model = model_binary(data, formula, weights=rebalance_weights)
            prob_T = model.predict(data)
        else:
            prob_T = np.ones(len(data)) * np.average(
                data[self.treatment], weights=rebalance_weights
            )

        indices_T0 = data.index[data[self.treatment] == 0]
        prob_T[indices_T0] = 1 - prob_T[indices_T0]

        # compute nested IPW estimate
        indices = data[self.treatment] == assignment
        return np.mean((indices / prob_T) * Y)

    def _augmented_nested_ipw(
        self, data, assignment, model_binary=None, model_continuous=None
    ):
        """
        Augmented nested IPW estimator for the counterfactual mean E[Y(t)].

        :param data: pandas data frame containing the data.
        :param assignment: assignment value for treatment.
        :param model_binary: string specifying modeling strategy to use for binary variables: e.g. glm-binary.
        :param model_continuous: string specifying modeling strategy to use for continuous variables: e.g. glm-continuous.
        :return: float corresponding to computed E[Y(t)].
        """

        if self.strategy == "Not ID":
            raise RuntimeError(
                "Nested IPW will not return valid estimates as causal effect is not identified"
            )

        # get the rebalancing weights
        rebalance_weights = self._get_nested_rebalanced_weights(
            data, model_binary, model_continuous
        )

        # extract the outcome and get Markov pillow of the treatment
        Y = data[self.outcome]
        mp_T = self.graph.markov_pillow([self.treatment], self.n_order)

        if len(mp_T) != 0:
            # fit T | mp(T) and predict treatment probabilities
            formula_T = self.treatment + " ~ " + "+".join(mp_T)  # + "+ ones"
            model = model_binary(data, formula_T, weights=rebalance_weights)
            prob_T = model.predict(data)
            formula_Y = (
                self.outcome + " ~ " + self.treatment + "+" + "+".join(mp_T)
            )
        else:
            prob_T = np.ones(len(data)) * np.average(
                data[self.treatment], weights=rebalance_weights
            )
            formula_Y = self.outcome + " ~ " + self.treatment

        indices_T0 = data.index[data[self.treatment] == 0]
        prob_T[indices_T0] = 1 - prob_T[indices_T0]
        indices = data[self.treatment] == assignment

        # fit Y | T=t, mp(T) and predict outcomes under assignment T=t
        data_assign = data.copy()
        data_assign[self.treatment] = assignment
        if self.state_space_map_[self.outcome] == "binary":
            model = model_binary(data, formula_Y, weights=rebalance_weights)
        else:
            model = model_continuous(data, formula_Y, weights=rebalance_weights)
        Yhat_vec = model.predict(data_assign)

        # return ANIPW estimate
        return np.mean((indices / prob_T) * (Y - Yhat_vec) + Yhat_vec)

    def compute_effect(
        self,
        data,
        estimator,
        model_binary=None,
        model_continuous=None,
        n_bootstraps=0,
        alpha=0.05,
        report_log_odds=True,
    ):
        """
        Bootstrap functionality to compute the Average Causal Effect if the outcome is continuous
        or the Causal Odds Ratio if the outcome is binary. Returns the point estimate
        as well as lower and upper quantiles for a user specified confidence level.

        :param data: pandas data frame containing the data.
        :param estimator: string indicating what estimator to use: e.g. eff-apipw.
        :param model_binary: string specifying modeling strategy to use for binary variables: e.g. glm-binary.
        :param model_continuous: string specifying modeling strategy to use for continuous variables: e.g. glm-continuous.
        :param n_bootstraps: number of bootstraps.
        :param alpha: the significance level with the default value of 0.05.
        :param report_log_odds: if set to True, and the outcome variable is binary, then reports the log odds ratio.
        :return: one float corresponding to ACE/OR if n_bootstraps=0, else three floats corresponding to ACE/OR, lower quantile, upper quantile.
        """
        df = data.copy()

        # instantiate modeling strategy with defaults
        if not model_binary:
            model_binary = self._fit_binary_glm
        if not model_continuous:
            model_continuous = self._fit_continuous_glm

        # add a column of ones to fit intercept terms
        df.loc[:, "ones"] = np.ones(len(df))

        # get state space of all variables
        for colname, colvalues in df.iteritems():
            if set([0, 1]).issuperset(colvalues.unique()):
                self.state_space_map_[colname] = "binary"
            else:
                self.state_space_map_[colname] = "continuous"

        # instantiate estimator and get point estimate of ACE
        method = self.estimators[estimator]
        point_estimate_T1 = method(df, 1, model_binary, model_continuous)
        point_estimate_T0 = method(df, 0, model_binary, model_continuous)

        # if Y is binary report log of odds ration, if Y is continuous report ACE
        if report_log_odds and self.state_space_map_[self.outcome] == "binary":
            ace = np.log(
                (point_estimate_T1 / (1 - point_estimate_T1))
                / (point_estimate_T0 / (1 - point_estimate_T0))
            )
        else:
            ace = point_estimate_T1 - point_estimate_T0

        if n_bootstraps > 0:
            ace_vec = []

            Ql = alpha / 2
            Qu = 1 - alpha / 2

            # iterate over bootstraps
            for iter in range(n_bootstraps):

                # resample the data with replacement
                df_sampled = df.sample(len(df), replace=True)
                df_sampled.reset_index(drop=True, inplace=True)

                # estimate ACE in resampled data
                estimate_T1 = method(
                    df_sampled, 1, model_binary, model_continuous
                )
                estimate_T0 = method(
                    df_sampled, 0, model_binary, model_continuous
                )

                # if Y is binary report log of odds ration, if Y is continuous report ACE
                if self.state_space_map_[self.outcome] == "binary":
                    ace_vec.append(
                        np.log(
                            (estimate_T1 / (1 - estimate_T1))
                            / (estimate_T0 / (1 - estimate_T0))
                        )
                    )
                else:
                    ace_vec.append(estimate_T1 - estimate_T0)

            # calculate the quantiles
            quantiles = np.quantile(ace_vec, q=[Ql, Qu])
            q_low = quantiles[0]
            q_up = quantiles[1]
            return ace, q_low, q_up

        return ace



import copy
import itertools
from collections import ChainMap

import numpy as np
import pytest
import sympy as sp

from ananke.estimation import empirical_plugin
from ananke.graphs import ADMG, DAG
from ananke.identification import OneLineID, one_line, oracle
from ananke.inference.variable_elimination import variable_elimination
from ananke.models import bayesian_network


class TestEffectEstimationID:
    def test_estimate_effect_from_distribution(self):
        di_edges = [("A", "Y"), ("C", "A"), ("C", "Y")]
        graph = ADMG(["A", "C", "Y"], di_edges=di_edges)
        cpds = bayesian_network.generate_random_cpds(graph)
        net = bayesian_network.BayesianNetwork(graph, cpds)
        treatment_dict = {"A": 1}
        outcome_dict = {"Y": 1}
        int_net = copy.deepcopy(net)
        int_net.fix(treatment_dict)

        int_inference = variable_elimination(int_net, ["Y"])
        truth = int_inference.get_value(**outcome_dict)

        oid = OneLineID(graph, list(treatment_dict), list(outcome_dict))
        obs_dist = empirical_plugin.get_obs_dist_from_net(net, graph)
        effect = empirical_plugin.estimate_effect_from_discrete_dist(
            oid, obs_dist, treatment_dict, outcome_dict
        )
        assert truth == pytest.approx(effect)

    def test_estimate_effect_front_door(self):
        di_edges = [("A", "M"), ("M", "Y")]
        bi_edges = [("A", "Y")]
        graph = ADMG(
            vertices={"A": 2, "M": 2, "Y": 2},
            di_edges=di_edges,
            bi_edges=bi_edges,
        )
        # TEST FAILING DUE TO BAD CANONICAL DAG CODE
        dag = graph.canonical_dag(cardinality=2)
        cpds = bayesian_network.generate_random_cpds(graph=dag, dir_conc=10)
        net = bayesian_network.BayesianNetwork(graph=dag, cpds=cpds)
        obs_dist = variable_elimination(net, list(graph.vertices))
        # VariableElimination(net).query(list(graph.vertices))
        treatment_dict = {"A": 1}

        outcome_dict = {"Y": 1}

        truth = oracle.compute_effect_from_discrete_model(
            net, treatment_dict, outcome_dict
        )
        oid = OneLineID(graph, list(treatment_dict), list(outcome_dict))
        effect = empirical_plugin.estimate_effect_from_discrete_dist(
            oid, obs_dist, treatment_dict, outcome_dict
        )

        assert truth == pytest.approx(effect)

    def test_effect_estimation_from_symbolic_distribution(self):

        di_edges = [("A", "M"), ("M", "Y")]
        bi_edges = [("A", "Y")]
        graph = ADMG(
            vertices={"A": 2, "M": 2, "Y": 2},
            di_edges=di_edges,
            bi_edges=bi_edges,
        )
        dag = graph.canonical_dag(cardinality=2)
        cpds, all_vars = bayesian_network.create_symbolic_cpds(dag)
        net = bayesian_network.BayesianNetwork(graph=dag, cpds=cpds)
        treatment_dict = {"A": 1}

        outcome_dict = {"Y": 1}
        truth = oracle.compute_effect_from_discrete_model(
            net, treatment_dict, outcome_dict
        )
        oid = OneLineID(graph, list(treatment_dict), list(outcome_dict))
        obs_dist = variable_elimination(net, list(graph.vertices))
        effect = empirical_plugin.estimate_effect_from_discrete_dist(
            oid, obs_dist, treatment_dict, outcome_dict
        )

        assert sp.simplify(truth - effect) == 0

        pass



from .admg import ADMG
from .bg import BG
from .cg import CG
from .dag import DAG
from .graph import *
from .ig import IG
from .missing_admg import MissingADMG
from .sg import SG
from .ug import UG
from .vertex import Vertex



import unittest

import ananke
from ananke.graphs import DAG


class TestDAG(unittest.TestCase):
    def test_dag_can_be_initialised(self):
        di_edges = [("A", "B")]
        vertices = ["A", "B"]
        dag = DAG(vertices, di_edges=di_edges)
        self.assertTrue(DAG)
        # test empty DAG
        DAG()

    def test_dseparation_Y_graph(self):
        vertices = ["X", "Y", "Z", "D"]
        di_edges = [("X", "Y"), ("Z", "Y"), ("Y", "D")]
        G = DAG(vertices, di_edges)
        self.assertFalse(G._d_separated("X", "Z", ["D"]))

    def test_dseparation_W_graph(self):
        vertices = ["A", "B", "C", "D", "E", "F", "G", "H"]
        di_edges = [
            ("A", "B"),
            ("C", "B"),
            ("C", "E"),
            ("D", "E"),
            ("B", "F"),
            ("F", "G"),
            ("F", "H"),
        ]
        G2 = DAG(vertices, di_edges)

        self.assertFalse(G2._d_separated("A", "D", ["F", "E"]))
        self.assertTrue(G2._d_separated("A", "H", ["B"]))
        self.assertFalse(G2._d_separated("A", "C", ["H"]))

    def test_d_separation_M_graph(self):
        vertices = ["Q", "W", "E", "R", "T", "I", "O", "U", "Y"]
        di_edges = [
            ("Q", "W"),
            ("Q", "E"),
            ("E", "R"),
            ("E", "T"),
            ("T", "U"),
            ("T", "Y"),
            ("I", "T"),
            ("I", "O"),
        ]
        G3 = DAG(vertices, di_edges)

        self.assertFalse(G3._d_separated("Q", "R", ["Y"]))
        self.assertTrue(G3._d_separated("Q", "Y", ["E"]))
        self.assertFalse(G3._d_separated("E", "R", ["I", "O"]))

    def test_d_separation_sets(self):
        vertices = ["A", "B", "C", "D"]
        di_edges = [("A", "B"), ("B", "C"), ("C", "D")]
        G = DAG(vertices, di_edges)

        self.assertTrue(G.d_separated("A", ["C", "D"], "B"))
        self.assertFalse(G.d_separated("A", ["C", "D"], []))

    def test_d_separation_fixed_and_random(self):
        vertices = ["A", "B", "C", "D"]
        di_edges = [("A", "B"), ("B", "C"), ("C", "D")]

        G = DAG(vertices, di_edges)
        G.fix("B")

        self.assertTrue(G.d_separated("B", ["A", "D"], ["C"]))

    def test_d_separation_fixed_vertices_only(self):
        vertices = ["A", "B", "C", "D"]
        di_edges = [("A", "B"), ("B", "C"), ("C", "D")]

        G = DAG(vertices, di_edges)
        G.fix(["A", "B"])

        with self.assertRaises(ananke.graphs.dag.UndefinedDAGOperation) as f:
            G.d_separated("B", ["A"], ["C"])

        print(str(f.exception))

    def test_subgraph_is_a_dag(self):
        vertices = ["A", "B", "C", "D"]
        di_edges = [("A", "B"), ("B", "C"), ("C", "D")]

        G = DAG(vertices, di_edges)
        G.fix(["A", "B"])

        subgraph = G.subgraph(["A", "B"])
        assert isinstance(subgraph, DAG)


if __name__ == "__main__":
    unittest.main()



"""
Implementation of DiscreteFator.

API inspired by pgmpy (https://github.com/pgmpy)
"""
import copy
import math

import numpy as np
import sympy as sp


class BaseDiscreteFactor:
    """
    The `inplace=False` argument is to maintain API compatibility with
    `pgmpy.factors.discrete.DiscreteFactor`.

    """

    def __init__(self, variables, cardinality, values):
        pass

    def get_cards_dict(self):
        return {k: v for k, v in zip(self.variables, self.cardinality)}

    def marginalize(self, other, inplace=False):
        raise NotImplementedError

    def product(self, other, inplace=False):
        raise NotImplementedError

    def reduce(self, values, inplace=False):
        raise NotImplementedError

    def divide(self, other, inplace=False):
        raise NotImplementedError


class SymDiscreteFactor(BaseDiscreteFactor):
    def __init__(self, variables, cardinality, values):
        """
        Initializes a symbolic `DiscreteFactor`, which is backed by sympy arrays.



        :param variables: Scope of the factor
        :type variables: list
        :param cardinality: Cardinality of each variable of the factor, in order supplied by
        `variables`
        :type cardinality: list
        :param values: A sympy.Array of variables representing the factor, or sympy.One
        :type values: sympy.Array, int

        """
        self.variables = list(variables)
        self.cardinality = cardinality
        if hasattr(values, "reshape"):
            self.values = values.reshape(*tuple(self.cardinality))
        else:
            self.values = values

    def marginalize(self, variables, inplace=False):
        if not variables:
            return copy.deepcopy(self)

        var_indexes = [self.variables.index(var) for var in variables]
        index_to_keep = sorted(
            set(range(len(self.variables))) - set(var_indexes)
        )

        new_values = sp.tensorcontraction(self.values, tuple(var_indexes))

        return SymDiscreteFactor(
            variables=[self.variables[i] for i in index_to_keep],
            cardinality=[self.cardinality[i] for i in index_to_keep],
            values=new_values,
        )

    def product(self, other, inplace=False):
        common_vars = sorted(set(self.variables) & set(other.variables))
        result = sp.tensorproduct(self.values, other.values)
        named_indexes = list(self.variables) + list(other.variables)
        if common_vars:
            for c in common_vars:
                active_indices = [
                    i for i, x in enumerate(named_indexes) if x == c
                ]
                result = sp.tensordiagonal(result, active_indices)
                named_indexes = [i for i in named_indexes if i != c] + [c]
        cards = dict(
            zip(
                self.variables + other.variables,
                self.cardinality + other.cardinality,
            )
        )
        final_cardinality = [cards[x] for x in named_indexes]

        return SymDiscreteFactor(
            variables=named_indexes,
            cardinality=final_cardinality,
            values=result.reshape(math.prod(final_cardinality)),
        )

    def reduce(self, evidence, inplace=False):
        slice_ = [slice(None)] * len(self.variables)
        var_index_to_del = list()
        for v, s in evidence:
            var_index = self.variables.index(v)
            slice_[var_index] = s
            var_index_to_del.append(var_index)
        var_index_to_keep = sorted(
            set(range(len(self.variables))) - set(var_index_to_del)
        )
        new_variables = [self.variables[i] for i in var_index_to_keep]
        new_cardinality = [self.cardinality[i] for i in var_index_to_keep]
        new_values = self.values[tuple(slice_)]

        return SymDiscreteFactor(
            variables=new_variables,
            cardinality=new_cardinality,
            values=new_values,
        )

    def divide(self, other, inplace=False):
        inverted_other = SymDiscreteFactor(
            variables=other.variables,
            cardinality=other.cardinality,
            values=sp.Array(1 / np.array(other.values)),
        )

        return self.product(inverted_other)

    def get_value(self, **kwargs):
        assert set(kwargs.keys()) == set(
            self.variables
        ), "Factor variables do not match specified variables"

        ix = [kwargs[var] for var in self.variables]
        return self.values[tuple(ix)]

    def __eq__(self, other):
        if not (
            isinstance(self, SymDiscreteFactor)
            and isinstance(other, SymDiscreteFactor)
        ):
            return False
        elif set(self.variables) != set(other.variables):
            return False
        else:
            if self.variables != other.variables:

                other_indexes = [
                    other.variables.index(var) for var in self.variables
                ]
                other_values = sp.permutedims(other.values, other_indexes)
                phi = SymDiscreteFactor(
                    variables=[other.variables[i] for i in other_indexes],
                    cardinality=[other.cardinality[i] for i in other_indexes],
                    values=other_values,
                )
            else:
                phi = copy.deepcopy(other)
            if hasattr(self.values, "shape") and hasattr(phi.values, "shape"):
                if self.values.shape != phi.values.shape:
                    return False
            else:
                if self.values != phi.values:
                    return False

            if self.cardinality != phi.cardinality:
                return False

            return True

    def subs(self, vals):
        new_values = self.values.subs(vals)
        return SymDiscreteFactor(
            variables=self.variables,
            cardinality=self.cardinality,
            values=new_values,
        )

    def to_pgmpy(self):
        import pgmpy

        try:
            factor = pgmpy.factors.discrete.DiscreteFactor(
                variables=self.variables,
                cardinality=self.cardinality,
                values=np.array(self.values).astype(float64),
            )
        except TypeError:
            raise ValueError(
                "There are unsubstituted Sympy variables - cannot convert to pgmpy"
            )
        return factor


class SymCPD(SymDiscreteFactor):
    def __init__(
        self,
        variable: str,
        variable_card: int,
        values: sp.Array,
        evidence=None,
        evidence_card=None,
    ):

        variables = [variable]
        cardinality = [variable_card]
        if evidence is not None:
            variables.extend(evidence)
            cardinality.extend(evidence_card)
        self.variable = variable
        self.variable_card = variable_card
        super().__init__(variables, cardinality, values)

    def get_values(self):
        if len(self.variables) > 1:
            return self.values.reshape(
                self.variable_card, math.prod(self.cardinality[1:])
            )
        else:
            return self.values.reshape(self.variable_card, 1)

    def to_factor(self):
        return SymDiscreteFactor(
            self.variables.copy(),
            self.cardinality.copy(),
            copy.deepcopy(self.values),
        )

    def to_pgmpy(self):
        from pgmpy.factors.discrete import TabularCPD

        values = self.get_values()

        np_values = np.array(values.tolist()).astype(float)

        try:
            if len(self.variables) == 1:
                factor = TabularCPD(
                    variable=self.variables[0],
                    variable_card=self.cardinality[0],
                    values=np_values,
                )
            else:
                factor = TabularCPD(
                    variable=self.variables[0],
                    variable_card=self.cardinality[0],
                    values=np_values,
                    evidence=self.variables[1:],
                    evidence_card=self.cardinality[1:],
                )
        except TypeError as e:
            raise ValueError(
                f"There are unsubstituted Sympy variables - cannot convert to pgmpy: {self.values}, {e}"
            )
        return factor



"""
Class for missing data acyclic directed mixed graphs
"""

from .admg import ADMG


class MissingADMG(ADMG):
    def __init__(self, vertices=[], di_edges=set(), bi_edges=set(), **kwargs):
        """
        Constructor

        Naming convention for vertices: R_name --> missingness indicator, X_name --> counterfactual
        Xp_name are proxies.

        :param vertices:
        :param di_edges:
        :param bi_edges:
        :param kwargs:
        """

        # initialize vertices in ADMG
        super().__init__(
            vertices=vertices, di_edges=di_edges, bi_edges=bi_edges, **kwargs
        )

        # missing data specific stuff
        self.counterfactuals = [v for v in self.vertices if v.startswith("X_")]
        self.indicators = [v for v in self.vertices if v.startswith("R_")]
        self.proxies = [v for v in self.vertices if v.startswith("Xp_")]

        for v in self.counterfactuals:
            if v.startswith("X_"):
                var_name = v.split("_")[1]
                self.add_diedge(v, "Xp_" + var_name)
                self.add_diedge("R_" + var_name, "Xp_" + var_name)

    def draw(self, direction=None):
        """
        Visualize the graph.

        :return : dot language representation of the graph.
        """
        from graphviz import Digraph

        dot = Digraph()

        # set direction from left to right if that's preferred
        if direction == "LR":
            dot.graph_attr["rankdir"] = direction

        for v in self.vertices.values():
            dot.node(
                v.name,
                shape="square" if v.fixed else "plaintext",
                height=".5",
                width=".5",
            )

        for parent, child in self.di_edges:
            # special clause for SWIGs
            if child.startswith("Xp_"):
                dot.edge(parent, child, color="grey")
            else:
                dot.edge(parent, child, color="blue")
        for sib1, sib2 in self.bi_edges:
            dot.edge(sib1, sib2, dir="both", color="red")

        return dot



import unittest

from ananke.datasets import helpers


class TestCounterfactualMean(unittest.TestCase):
    def test_loaders(self):

        data = helpers.load_conditionally_ignorable_data()
        self.assertTrue(len(data) == 2000)
        self.assertTrue("ViralLoad" in data)

        data = helpers.load_afixable_data()
        self.assertTrue(len(data) == 2000)
        self.assertTrue("Education" in data)

        data = helpers.load_frontdoor_data()
        self.assertTrue(len(data) == 2000)
        self.assertTrue("ViralLoad2" in data)


if __name__ == "__main__":
    unittest.main()



"""
Class for Bidirected Graphs (BGs).
"""


from .admg import ADMG


class BG(ADMG):
    def __init__(self, vertices=[], bi_edges=set(), **kwargs):
        """
        Constructor.

        :param vertices: iterable of names of vertices.
        :param bi_edges: iterable of tuples of bidirected edges i.e. (X, Y) = X <-> Y.
        """

        super().__init__(vertices=vertices, bi_edges=bi_edges, **kwargs)



import copy
import functools
import logging
from collections import ChainMap

import numpy as np
from pgmpy.factors.discrete import DiscreteFactor, TabularCPD
from pgmpy.inference import VariableElimination
from pgmpy.models import BayesianNetwork

from ananke.inference.variable_elimination import variable_elimination

from ..graphs import ADMG, DAG

logger = logging.getLogger(__name__)


def compute_effect_from_discrete_model(
    net, treatment_dict, outcome_dict, conditioning_dict=None
):
    """
    Compute the causal effect by directly performing an intervention in a Bayesian
    Network corresponding to the true structural equation model to obtain the
    counterfactual distribution, and then computing the marginal distribution of the outcome.
    Note that this function does not consider issues of identification as
    interventions are performed in the true model (regardless if those
    interventions were identified).

    :param net: A Bayesian Network representing the causal problem. Note that this object is used only as a representation of the observed data distribution.
    :param treatment_dict: Dictionary of treatment variables to treatment values.
    :param outcome_dict: Dictionary of outcome variables to outcome values.
    """

    int_net = copy.deepcopy(net)
    int_net.fix(treatment_dict)

    if conditioning_dict is None:
        truth = variable_elimination(
            int_net, list(outcome_dict.keys())
        ).get_value(**outcome_dict)
    else:
        num = variable_elimination(
            int_net, list(outcome_dict.keys()) + list(conditioning_dict.keys())
        )
        denom = variable_elimination(int_net, list(conditioning_dict.keys()))
        final = num.divide(denom, inplace=False)
        truth = final.get_value(**(outcome_dict | conditioning_dict))

    return truth
